@@ -5,19 +5,24 @@ import io
 import sys
 import tempfile
 import functools
+import os
 
-from visidata import VisiData, vd, asyncthread
-from visidata import Sheet, Path
+from visidata import VisiData, vd, asyncthread, SettableColumn
+from visidata import Sheet, Path, Column
 
 if sys.platform == 'win32':
-    syscopy_cmd_default = 'clip'
-    syspaste_cmd_default = 'clip'
+    syscopy_cmd_default = 'clip.exe'
+    syspaste_cmd_default = 'powershell -command Get-Clipboard'
 elif sys.platform == 'darwin':
     syscopy_cmd_default = 'pbcopy w'
     syspaste_cmd_default = 'pbpaste'
 else:
-    syscopy_cmd_default = 'xclip -selection clipboard -filter'  # xsel --clipboard --input
-    syspaste_cmd_default = 'xclip -selection clipboard -o'  # xsel --clipboard
+    if 'WAYLAND_DISPLAY' in os.environ:
+        syscopy_cmd_default = 'wl-copy'
+        syspaste_cmd_default = 'wl-paste'
+    else:
+        syscopy_cmd_default = 'xclip -selection clipboard -filter'  # xsel --clipboard --input
+        syspaste_cmd_default = 'xclip -selection clipboard -o'  # xsel --clipboard
 
 vd.option('clipboard_copy_cmd', syscopy_cmd_default, 'command to copy stdin to system clipboard', sheettype=None)
 vd.option('clipboard_paste_cmd', syspaste_cmd_default, 'command to send contents of system clipboard to stdout', sheettype=None)
@@ -48,7 +53,7 @@ def syscopyValue(sheet, val):
     p = subprocess.run(
         sheet.options.clipboard_copy_cmd.split(),
         input=val,
-        encoding=sheet.options.encoding,
+        encoding='utf-8',
         stdout=subprocess.DEVNULL)
 
     vd.status('copied value to system clipboard')
@@ -70,16 +75,19 @@ def syscopyCells_async(sheet, cols, rows, filetype):
     vd.status(f'copying {vs.nRows} {vs.rowtype} to system clipboard as {filetype}')
 
     with io.StringIO() as buf:
-        vd.sync(vd.saveSheets(Path(sheet.name+'.'+filetype, fptext=buf), vs))
-        subprocess.run(
-            sheet.options.clipboard_copy_cmd.split(),
-            input=buf.getvalue(),
-            encoding=sheet.options.encoding,
-            stdout=subprocess.DEVNULL)
+        with tempfile.NamedTemporaryFile() as temp:
+            temp.close()  #2118
+
+            vd.sync(vd.saveSheets(Path(f'{temp.name}.{filetype}', fptext=buf), vs, confirm_overwrite=False))
+            subprocess.run(
+                sheet.options.clipboard_copy_cmd.split(),
+                input=buf.getvalue(),
+                encoding='utf-8',
+                stdout=subprocess.DEVNULL)
 
 
 @VisiData.api
-def sysclip_value(vd):
+def sysclipValue(vd):
     cmd = vd.options.clipboard_paste_cmd
     return subprocess.check_output(vd.options.clipboard_paste_cmd.split()).decode('utf-8')
 
@@ -87,7 +95,7 @@ def sysclip_value(vd):
 @VisiData.api
 @asyncthread
 def pasteFromClipboard(vd, cols, rows):
-    text = vd.getLastArgs() or vd.sysclip_value().strip() or vd.fail('system clipboard is empty')
+    text = vd.getLastArgs() or vd.sysclipValue().strip() or vd.fail('system clipboard is empty')
 
     vd.addUndoSetValues(cols, rows)
     lines = text.split('\n')
@@ -108,6 +116,8 @@ def pasteFromClipboard(vd, cols, rows):
 
 @Sheet.api
 def delete_row(sheet, rowidx):
+    if not sheet.rows:
+        vd.fail("no row to delete")
     if not sheet.defer:
         oldrow = sheet.rows.pop(rowidx)
         vd.addUndo(sheet.rows.insert, rowidx, oldrow)
@@ -122,11 +132,33 @@ def delete_row(sheet, rowidx):
     sheet.setModified()
     return oldrow
 
-@Sheet.api
-def paste_after(sheet, rowidx):
-    to_paste = list(deepcopy(r) for r in reversed(vd.memory.cliprows))
-    sheet.addRows(to_paste, index=rowidx)
 
+@Sheet.api
+@asyncthread
+def paste_after(sheet, rowidx):
+    'Paste rows from *vd.cliprows* at *rowidx*.'
+    if not vd.memory.cliprows:  #1793
+        vd.warning('nothing to paste')
+        return
+
+    for col in vd.memory.clipcols[sheet.nVisibleCols:]:
+        newcol = SettableColumn()
+        newcol.__setstate__(col.__getstate__())
+        sheet.addColumn(newcol)
+
+    addedRows = []
+
+    for extrow in vd.memory.cliprows:
+        if isinstance(extrow, Column):
+            newrow = copy(extrow)
+        else:
+            newrow = sheet.newRow()
+            for col, extcol in zip(sheet.visibleCols, vd.memory.clipcols):
+                col.setValue(newrow, extcol.getTypedValue(extrow))
+
+        addedRows.append(newrow)
+
+    sheet.addRows(addedRows, index=rowidx)
 
 
 Sheet.addCommand('y', 'copy-row', 'copyRows([cursorRow])', 'yank (copy) current row to clipboard')
@@ -167,3 +199,27 @@ Sheet.addCommand('gzx', 'cut-cells', 'copyCells(cursorCol, onlySelectedRows); cu
 
 Sheet.bindkey('KEY_DC', 'delete-cell'),
 Sheet.bindkey('gKEY_DC', 'delete-cells'),
+
+vd.addMenuItems('''
+    Edit > Delete > current row > delete-row
+    Edit > Delete > current cell > delete-cell
+    Edit > Delete > selected rows > delete-selected
+    Edit > Delete > selected cells > delete-cells
+    Edit > Copy > current cell > copy-cell
+    Edit > Copy > current row > copy-row
+    Edit > Copy > selected cells > copy-cells
+    Edit > Copy > selected rows > copy-selected
+    Edit > Copy > to system clipboard > current cell > syscopy-cell
+    Edit > Copy > to system clipboard > current row > syscopy-row
+    Edit > Copy > to system clipboard > selected cells > syscopy-cells
+    Edit > Copy > to system clipboard > selected rows > syscopy-selected
+    Edit > Cut > current row > cut-row
+    Edit > Cut > selected cells > cut-selected
+    Edit > Cut > current cell > cut-cell
+    Edit > Paste > row after > paste-after
+    Edit > Paste > row before > paste-before
+    Edit > Paste > into selected cells > setcol-clipboard
+    Edit > Paste > into current cell > paste-cell
+    Edit > Paste > from system clipboard > cells at cursor > syspaste-cells
+    Edit > Paste > from system clipboard > selected cells > syspaste-cells-selected
+''')
