@@ -3,13 +3,26 @@ import urllib.parse
 import copy
 import itertools
 
-from visidata import VisiData, vd, Sheet, options, Column, Progress, IndexSheet, ItemColumn, InferColumnsSheet
+from visidata import VisiData, vd, Sheet, options, Column, Progress, IndexSheet, ItemColumn
 
 vd.option('html_title', '<h2>{sheet.name}</h2>', 'table header when saving to html')
 
+
+@VisiData.api
+def guess_html(vd, p):
+    with p.open() as fp:
+        r = fp.read(10240)
+        if r.strip().startswith('<'):
+            m = re.search(r, r'charset=(\S+)')
+            if m:
+                encoding = m.group(0)
+            else:
+                encoding = None
+            return dict(filetype='html', _likelihood=1, encoding=encoding)
+
 @VisiData.api
 def open_html(vd, p):
-    return HtmlTablesSheet(p.name, source=p)
+    return HtmlTablesSheet(p.base_stem, source=p)
 
 VisiData.open_htm = VisiData.open_html
 
@@ -20,12 +33,17 @@ class HtmlTablesSheet(IndexSheet):
         Column('tag', width=0, getter=lambda col,row: row.html.tag),
         Column('id', getter=lambda col,row: row.html.attrib.get('id')),
         Column('classes', getter=lambda col,row: row.html.attrib.get('class')),
+        Column('title', getter=lambda col,row: row.html.attrib.get('title')),
+        Column('aria_label', getter=lambda col,row: row.html.attrib.get('aria-label')),
+        Column('caption', getter=lambda col,row: row.html.xpath('normalize-space(./caption)') if row.html.xpath('./caption') else None, cache=True),
+        Column('summary', getter=lambda col,row: row.html.attrib.get('summary')),
+        Column('heading', getter=lambda col,row: row.html.xpath('normalize-space(./preceding-sibling::*[self::h1 or self::h2 or self::h3 or self::h4 or self::h5 or self::h6][1])') or None, cache=True),
     ]
     def iterload(self):
+        lxml = vd.importExternal('lxml')
         from lxml import html
-        utf8_parser = html.HTMLParser(encoding='utf-8')
-        with self.source.open_text(encoding='utf-8') as fp:
-            doc = html.parse(fp, parser=utf8_parser, base_url=self.source.given)
+        with self.source.open(encoding='utf-8') as fp:
+            doc = html.parse(fp, parser=vd.utf8_parser, base_url=self.source.given)
         self.setKeys([self.column('name')])
         self.column('keys').hide()
         self.column('source').hide()
@@ -56,15 +74,16 @@ class HtmlLinksSheet(Sheet):
         ItemColumn('link', 2, width=40),
     ]
     def iterload(self):
+        lxml = vd.importExternal('lxml')
         from lxml.html import iterlinks
         root = self.source.getroot()
-        root.make_links_absolute(self.source.docinfo.URL)
+        root.make_links_absolute(self.source.docinfo.URL, handle_failures='ignore')
         yield from iterlinks(root)
 
     def openRow(self, row):
         return vd.openSource(row[2])
 
-class HtmlElementsSheet(InferColumnsSheet):
+class HtmlElementsSheet(Sheet):
     rowtype = 'links'  #  dict
     columns = [
         ItemColumn('__element__', width=0),
@@ -89,6 +108,7 @@ class HtmlTableSheet(Sheet):
     columns = []
 
     def iterload(self):
+        import lxml
         headers = []
 
         maxlinks = {}  # [colnum] -> nlinks:int
@@ -108,8 +128,14 @@ class HtmlTableSheet(Sheet):
             for cell in r.getchildren():
                 colspan = int(cell.attrib.get('colspan', 1))
                 rowspan = int(cell.attrib.get('rowspan', 1))
+                if isinstance(cell, lxml.etree.CommentBase):
+                    continue
                 cellval = ' '.join(x.strip() for x in cell.itertext())  # text only without markup
-                links = [urllib.parse.urljoin(self.source.base_url, x.get('href')) for x in cell.iter('a')]
+                links = [
+                    vd.callNoExceptions(urllib.parse.urljoin, self.source.base_url, x.get('href')) or x.get('href')
+                        for x in cell.iter('a')
+                ]
+
                 maxlinks[colnum] = max(maxlinks.get(colnum, 0), len(links))
 
                 if is_header(cell):
@@ -124,7 +150,7 @@ class HtmlTableSheet(Sheet):
                         cellval = ''   # use empty non-None value for subsequent rows in the rowspan
                 else:
                     while colnum >= len(row):
-                        row.append(None)
+                        row.append((None, []))
                     row[colnum] = (cellval, links)
 
                 colnum += colspan
@@ -137,8 +163,11 @@ class HtmlTableSheet(Sheet):
         if headers:
             it = itertools.zip_longest(*headers, fillvalue='')
         else:
-            it = list(list(x) for x in self.rows.pop(0))
-            it += [''] * (ncols-len(it))
+            if len(self.rows) > 0:
+                it = list(list(x) for x in self.rows.pop(0))
+                it += [''] * (ncols-len(it))
+            else:
+                it = []
 
         for colnum, names in enumerate(it):
             name = '_'.join(str(x) for x in names if x)
@@ -177,7 +206,20 @@ def save_html(vd, p, *vsheets):
                     fp.write('</tr>\n')
 
             fp.write('</table>')
-            vd.status('%s save finished' % p)
+
+
+@VisiData.lazy_property
+def utf8_parser(vd):
+    lxml = vd.importExternal('lxml')
+    return lxml.html.HTMLParser(encoding='utf-8')
+#    return lxml.etree.HTMLParser(encoding='utf-8')
+
+
+@VisiData.api
+def HTML(vd, s):
+    lxml = vd.importExternal('lxml')
+    from lxml import html
+    return html.fromstring(s, parser=vd.utf8_parser)
 
 
 VisiData.save_htm = VisiData.save_html
