@@ -3,7 +3,7 @@ from visidata import VisiData, vd, Sheet, Column, Progress, ColumnAttr, ColumnIt
 
 @VisiData.api
 def open_ttf(vd, p):
-    return TTFTablesSheet(p.name, source=p)
+    return TTFTablesSheet(p.base_stem, source=p)
 
 vd.open_otf = vd.open_ttf
 
@@ -24,9 +24,9 @@ class TTFTablesSheet(Sheet):
         return TTFGlyphsSheet(self.name+'_glyphs', source=self, sourceRows=[row], ttf=self.ttf)
 
     def iterload(self):
-        import fontTools.ttLib
+        fontTools = vd.importExternal('fontTools.ttLib', 'fonttools')
 
-        self.ttf = fontTools.ttLib.TTFont(str(self.source), 0, allowVID=0, ignoreDecompileErrors=True, fontNumber=-1)
+        self.ttf = fontTools.TTFont(str(self.source), 0, allowVID=0, ignoreDecompileErrors=True, fontNumber=-1)
         for cmap in self.ttf["cmap"].tables:
             yield cmap
 
@@ -53,20 +53,21 @@ class TTFGlyphsSheet(Sheet):
 
 
 def makePen(*args, **kwargs):
-    try:
-        from fontTools.pens.basePen import BasePen
-    except ImportError as e:
-        vd.error('fonttools not installed')
+    fontTools = vd.importExternal('fontTools', 'fonttools')
+    from fontTools.pens.basePen import BasePen
 
     class GlyphPen(InvertedCanvas, BasePen):
         aspectRatio = 1.0
         def __init__(self, name, **kwargs):
             super().__init__(name, **kwargs)
+            self.path_firstxy = None
             self.lastxy = None
             self.attr = self.plotColor(('glyph',))
 
         def _moveTo(self, xy):
             self.lastxy = xy
+            if self.path_firstxy is None:
+                self.path_firstxy = xy
 
         def _lineTo(self, xy):
             x1, y1 = self.lastxy
@@ -74,6 +75,17 @@ def makePen(*args, **kwargs):
             self.line(x1, y1, x2, y2, self.attr)
             self._moveTo(xy)
 
+        def _closePath(self):
+            if self.path_firstxy:
+                if (self.path_firstxy != self.lastxy):
+                    self._lineTo(self.path_firstxy)
+                self.path_firstxy = None
+            self.lastxy = None
+
+        def _endPath(self):
+            self.path_firstxy = None
+            self.lastxy = None
+
         def _curveToOne(self, xy1, xy2, xy3):
             vd.error('NotImplemented')
 
