@@ -1,9 +1,10 @@
-from functools import singledispatch
 from typing import Mapping
 import inspect
 import math
 
-from visidata import *
+from visidata import vd, asyncthread, ENTER, deduceType
+from visidata import Sheet, Column, VisiData, ColumnItem, TableSheet, BaseSheet, Progress, ColumnAttr, SuspendCurses, TextSheet
+import visidata
 
 vd.option('visibility', 0, 'visibility level')
 vd.option('default_sample_size', 100, 'number of rows to sample for regex.split (0=all)', replay=True)
@@ -16,177 +17,6 @@ class PythonSheet(Sheet):
         return PyobjSheet("%s[%s]" % (self.name, self.keystr(row)), source=row)
 
 
-class InferColumnsSheet(Sheet):
-    _rowtype = dict
-    @asyncthread
-    def reload(self):
-        self.reloadCols()
-
-        self.rows = []
-        for r in self.iterload():
-            self.addRow(r)
-
-        # if an ordering has been specified, sort the sheet
-        if self._ordering:
-            vd.sync(self.sort())
-
-    def reloadCols(self):
-        self.columns = []
-        self._knownKeys.clear()
-        for c in type(self).columns:
-            self.addColumn(deepcopy(c))
-
-    def addColumn(self, *cols, index=None):
-        for c in cols:
-            self._knownKeys.add(c.expr or c.name)
-        return super().addColumn(*cols, index=index)
-
-    def addRow(self, row, index=None):
-        ret = super().addRow(row, index=index)
-        for k in row:
-            if k not in self._knownKeys:
-                self.addColumn(ColumnItem(k, type=deduceType(row[k])))
-
-        return ret
-
-
-InferColumnsSheet.init('_knownKeys', set, copy=True)  # set of row keys already seen
-InferColumnsSheet.init('_ordering', list, copy=True)
-
-
-@Sheet.api
-def getSampleRows(sheet):
-    'Return list of sample rows centered around the cursor.'
-    n = sheet.options.default_sample_size
-    if n == 0 or n >= sheet.nRows:
-        return sheet.rows
-
-    vd.warning(f'sampling {n} rows')
-    seq = sheet.rows
-    start = math.ceil(sheet.cursorRowIndex - n / 2) % len(seq)
-    end = (start + n) % len(seq)
-    if start < end:
-        return seq[start:end]
-    return seq[start:] + seq[:end]
-
-
-@asyncthread
-def expand_cols_deep(sheet, cols, rows=None, depth=0):  # depth == 0 means drill all the way
-    'expand all visible columns of containers to the given depth (0=fully)'
-    ret = []
-    if not rows:
-        rows = sheet.getSampleRows()
-
-    for col in cols:
-        newcols = col.expand(rows)
-        if depth != 1:  # countdown not yet complete, or negative (indefinite)
-            ret.extend(expand_cols_deep.__wrapped__(sheet, newcols, rows, depth-1))
-    return ret
-
-@singledispatch
-def _createExpandedColumns(sampleValue, col, rows):
-    '''By default, a column is not expandable. Supported container types for
-    sampleValue trigger alternate, type-specific expansions.'''
-    return []
-
-@_createExpandedColumns.register(dict)
-def _(sampleValue, col, vals):
-    '''Build a set of columns to add, using the first occurrence of each key to
-    determine column type'''
-    newcols = {}
-
-    for val in Progress(vals, 'expanding'):
-        colsToAdd = set(val).difference(newcols)
-        colsToAdd and newcols.update({
-            k: deduceType(v)
-            for k, v in val.items()
-            if k in colsToAdd
-        })
-
-    return [
-        ExpandedColumn(col.sheet.options.fmt_expand_dict % (col.name, k), type=v, origCol=col, expr=k)
-            for k, v in newcols.items()
-    ]
-
-def _createExpandedColumnsNamedTuple(col, val):
-    return [
-        ExpandedColumn(col.sheet.options.fmt_expand_dict % (col.name, k), type=colType, origCol=col, expr=i)
-            for i, (k, colType) in enumerate(zip(val._fields, (deduceType(v) for v in val)))
-    ]
-
-@_createExpandedColumns.register(list)
-@_createExpandedColumns.register(tuple)
-def _(sampleValue, col, vals):
-    '''Use the longest sequence to determine the number of columns we need to
-    create, and their presumed types'''
-    def lenNoExceptions(v):
-        try:
-            return len(v)
-        except Exception as e:
-            return 0
-
-    if hasattr(sampleValue, '_fields'):  # looks like a namedtuple
-        return _createExpandedColumnsNamedTuple(col, vals[0])
-
-    longestSeq = max(vals, key=lenNoExceptions)
-    colTypes = [deduceType(v) for v in longestSeq]
-    return [
-        ExpandedColumn(col.sheet.options.fmt_expand_list % (col.name, k), type=colType, origCol=col, expr=k)
-            for k, colType in enumerate(colTypes)
-    ]
-
-
-@Column.api
-def expand(col, rows):
-    isNull = col.sheet.isNullFunc()
-    nonNulls = [
-        col.getTypedValue(row)
-        for row in rows
-        if not isNull(col.getValue(row))
-    ]
-
-    if not nonNulls:
-        return []
-
-    # The type of the first non-null value for col determines if and how the
-    # column can be expanded.
-    expandedCols = _createExpandedColumns(nonNulls[0], col, nonNulls)
-
-    idx = col.sheet.columns.index(col)
-
-    for i, c in enumerate(expandedCols):
-        col.sheet.addColumn(c, index=idx+i+1)
-    if expandedCols:
-        col.hide()
-    return expandedCols
-
-
-def deduceType(v):
-    if isinstance(v, (float, int)):
-        return type(v)
-    else:
-        return anytype
-
-
-class ExpandedColumn(Column):
-    def calcValue(self, row):
-        return getitemdef(self.origCol.getValue(row), self.expr)
-
-    def setValue(self, row, value):
-        self.origCol.getValue(row)[self.expr] = value
-
-
-def closeColumn(sheet, col):
-    if hasattr(col, 'origCol'):
-        origCol = col.origCol
-    else:
-        vd.fail('column has not been expanded')
-    vd.addUndo(setattr, sheet, 'columns', sheet.columns)
-    origCol.width = options.default_width
-    cols = [c for c in sheet.columns if getattr(c, "origCol", None) is not origCol]
-    sheet.columns = cols
-
-
 #### generic list/dict/object browsing
 @VisiData.global_api
 def view(vd, obj):
@@ -208,14 +38,14 @@ def AttrColumns(attrnames):
 
 
 def SheetList(*names, **kwargs):
-    'Creates a Sheet from a list of homogenous dicts or namedtuples.'
+    'Creates a Sheet from a list of homogeneous dicts or namedtuples.'
 
     src = kwargs.get('source', None)
     if not src:
-        vd.status('no content in %s' % names)
-        return
+        vd.warning('no content in %s' % names)
+        return Sheet(*names, **kwargs)
 
-    if isinstance(src[0], dict):
+    if isinstance(src[0], Mapping):
         return ListOfDictSheet(*names, **kwargs)
     elif isinstance(src[0], tuple):
         if getattr(src[0], '_fields', None):  # looks like a namedtuple
@@ -226,7 +56,7 @@ def SheetList(*names, **kwargs):
 
 class ListOfPyobjSheet(PythonSheet):
     rowtype = 'python objects'
-    def reload(self):
+    def loader(self):
         self.rows = self.source
         self.columns = []
         self.addColumn(Column(self.name,
@@ -275,7 +105,7 @@ class SheetNamedTuple(PythonSheet):
         self.rows = list(zip(self.source._fields, self.source))
 
     def openRow(self, row):
-        return PyobjSheet(self.name, row[0], source=row[1])
+        return PyobjSheet(f'{self.name}.{row[0]}', source=row[1])
 
 
 # source is dict
@@ -291,7 +121,7 @@ class SheetDict(PythonSheet):
         self.rows = list(self.source.keys())
 
     def openRow(self, row):
-        return PyobjSheet(self.name, row, source=self.source[row])
+        return PyobjSheet(f'{self.name}.{row}', source=self.source[row])
 
 
 class ColumnSourceAttr(Column):
@@ -347,10 +177,11 @@ class PyobjSheet(PythonSheet):
                 if vislevel <= 2 and r.startswith('__'): continue
                 if vislevel <= 1 and r.startswith('_'): continue
                 if vislevel <= 0 and callable(getattr(self.source, r)): continue
-                self.addRow(r)
             except Exception:
                 pass
 
+            self.addRow(r)
+
     def openRow(self, row):
         'dive further into Python object'
         v = getattr(self.source, row)
@@ -358,9 +189,13 @@ class PyobjSheet(PythonSheet):
 
 
 @TableSheet.api
-def openRow(sheet, row):
+def openRow(sheet, row, rowidx=None):
     'Return Sheet diving into *row*.'
-    k = sheet.keystr(row) or sheet.cursorRowIndex
+    if rowidx is None:
+        k = sheet.keystr(row) or str(sheet.cursorRowIndex)
+    else:
+        k = rowidx
+
     name = f'{sheet.name}[{k}]'
     return TableSheet(name,
                       rows=sheet.visibleCols,
@@ -372,41 +207,50 @@ def openRow(sheet, row):
                       nKeys=1)
 
 @TableSheet.api
-def openCell(sheet, col, row):
+def openCell(sheet, col, row, rowidx=None):
     'Return Sheet diving into cell at *row* in *col*.'
-    k = sheet.keystr(row) or [str(sheet.cursorRowIndex)]
+    if rowidx is None:
+        k = sheet.keystr(row) or str(sheet.cursorRowIndex)
+    else:
+        k = rowidx
     name = f'{sheet.name}[{k}].{col.name}'
     return PyobjSheet(name, source=col.getTypedValue(row))
 
+@TableSheet.api
+def openRowPyobj(sheet, rowidx):
+    'Return Sheet of raw Python object of row.'
+    return PyobjSheet("%s[%s]" % (sheet.name, rowidx), source=sheet.rows[rowidx])
+
+@TableSheet.api
+def openCellPyobj(sheet, col, rowidx):
+    'Return Sheet of raw Python object of cell.'
+    name = f'{sheet.name}[{rowidx}].{col.name}'
+    return PyobjSheet(name, source=col.getValue(sheet.rows[rowidx]))
+
 
 @BaseSheet.api
-def pyobj_expr(sheet):
+def inputPythonExpr(sheet):
     def launch_repl(v, i):
         import code
         with SuspendCurses():
             code.InteractiveConsole(locals=locals()).interact()
         return v, i
-    expr = vd.input("eval: ", "expr", completer=visidata.CompleteExpr(), bindings={'^X': launch_repl})
-    vd.push(PyobjSheet(expr, source=sheet.evalExpr(expr)))
+    return vd.input("eval: ", "expr", completer=visidata.CompleteExpr(), bindings={'^X': launch_repl})
 
-BaseSheet.addCommand('^X', 'pyobj-expr', 'pyobj_expr()', 'evaluate Python expression and open result as Python object')
+BaseSheet.addCommand('^X', 'pyobj-expr', 'expr=inputPythonExpr(); vd.push(PyobjSheet(expr, source=sheet.evalExpr(expr)))', 'evaluate Python expression and open result as Python object')
 BaseSheet.addCommand('', 'exec-python', 'expr = input("exec: ", "expr", completer=CompleteExpr()); exec(expr, getGlobals(), LazyChainMap(sheet, *vd.contexts, locals=vd.getGlobals()))', 'execute Python statement with expression scope')
 BaseSheet.addCommand('g^X', 'import-python', 'modname=input("import: ", type="import_python"); exec("import "+modname, getGlobals())', 'import Python module in the global scope')
-globalCommand('z^X', 'pyobj-expr-row', 'expr = input("eval over current row: ", "expr", completer=CompleteExpr()); vd.push(PyobjSheet(expr, source=evalExpr(expr, row=cursorRow)))', 'evaluate Python expression, in context of current row, and open result as Python object')
+BaseSheet.addCommand('z^X', 'pyobj-expr-row', 'expr = input("eval over current row: ", "expr", completer=CompleteExpr()); vd.push(PyobjSheet(expr, source=evalExpr(expr, row=cursorRow)))', 'evaluate Python expression, in context of current row, and open result as Python object')
+BaseSheet.addCommand('', 'assert-expr', 'expr=inputPythonExpr(); assert sheet.evalExpr(expr), f"{expr} not true"', 'eval Python expression and assert result is truthy')
+BaseSheet.addCommand('', 'assert-expr-row', 'expr=inputPythonExpr(); assert sheet.evalExpr(expr, row=cursorRow), f"{expr} not true"', 'eval Python expression in context of current row, and assert result is truthy')
 
-Sheet.addCommand('^Y', 'pyobj-row', 'status(type(cursorRow)); vd.push(PyobjSheet("%s[%s]" % (sheet.name, cursorRowIndex), source=cursorRow))', 'open current row as Python object')
-Sheet.addCommand('z^Y', 'pyobj-cell', 'status(type(cursorValue)); vd.push(PyobjSheet("%s[%s].%s" % (sheet.name, cursorRowIndex, cursorCol.name), source=cursorValue))', 'open current cell as Python object')
-globalCommand('g^Y', 'pyobj-sheet', 'status(type(sheet)); vd.push(PyobjSheet(sheet.name+"_sheet", source=sheet))', 'open current sheet as Python object')
+Sheet.addCommand('^Y', 'pyobj-row', 'status(type(cursorRow).__name__); vd.push(openRowPyobj(cursorRowIndex))', 'open current row as Python object')
+Sheet.addCommand('z^Y', 'pyobj-cell', 'status(type(cursorValue).__name__); vd.push(openCellPyobj(cursorCol, cursorRowIndex))', 'open current cell as Python object')
+BaseSheet.addCommand('g^Y', 'pyobj-sheet', 'status(type(sheet).__name__); vd.push(PyobjSheet(sheet.name+"_sheet", source=sheet))', 'open current sheet as Python object')
 
-Sheet.addCommand('(', 'expand-col', 'expand_cols_deep(sheet, [cursorCol], depth=1)', 'expand current column of containers one level')
-Sheet.addCommand('g(', 'expand-cols', 'expand_cols_deep(sheet, visibleCols, depth=1)', 'expand all visible columns of containers one level')
-Sheet.addCommand('z(', 'expand-col-depth', 'expand_cols_deep(sheet, [cursorCol], depth=int(input("expand depth=", value=1)))', 'expand current column of containers to given depth (0=fully)')
-Sheet.addCommand('gz(', 'expand-cols-depth', 'expand_cols_deep(sheet, visibleCols, depth=int(input("expand depth=", value=1)))', 'expand all visible columns of containers to given depth (0=fully)')
-
-Sheet.addCommand(')', 'contract-col', 'closeColumn(sheet, cursorCol)', 'unexpand current column; restore original column and remove other columns at this level')
-
-Sheet.addCommand(ENTER, 'open-row-pyobj', 'vd.push(openRow(cursorRow))', 'open sheet with copies of rows referenced in current row')
-Sheet.addCommand('z'+ENTER, 'open-cell-pyobj', 'vd.push(openCell(cursorCol, cursorRow))', 'open sheet with copies of rows referenced in current cell')
+Sheet.addCommand('', 'open-row-basic', 'vd.push(TableSheet.openRow(sheet, cursorRow))', 'dive into current row as basic table (ignoring subsheet dive)')
+Sheet.addCommand(ENTER, 'open-row', 'vd.push(openRow(cursorRow))', 'open current row with sheet-specific dive')
+Sheet.addCommand('z'+ENTER, 'open-cell', 'vd.push(openCell(cursorCol, cursorRow))', 'open sheet with copies of rows referenced in current cell')
 Sheet.addCommand('g'+ENTER, 'dive-selected', 'for r in selectedRows: vd.push(openRow(r))', 'open sheet with copies of rows referenced in selected rows')
 Sheet.addCommand('gz'+ENTER, 'dive-selected-cells', 'for r in selectedRows: vd.push(openCell(cursorCol, r))', 'open sheet with copies of rows referenced in selected rows')
 
@@ -415,14 +259,21 @@ PyobjSheet.addCommand('gv', 'show-hidden', 'sheet.options.visibility = 2; reload
 PyobjSheet.addCommand('zv', 'hide-hidden', 'sheet.options.visibility -= 1; reload()', 'hide methods and hidden properties')
 
 vd.addGlobals({
-    'ExpandedColumn': ExpandedColumn,
     'PythonSheet': PythonSheet,
-    'expand_cols_deep': expand_cols_deep,
-    'deduceType': deduceType,
-    'closeColumn': closeColumn,
     'ListOfDictSheet': ListOfDictSheet,
     'SheetDict': SheetDict,
-    'InferColumnsSheet': InferColumnsSheet,
     'PyobjSheet': PyobjSheet,
     'view': view,
 })
+
+vd.addMenuItems('''
+    View > Visibility > Methods and dunder attributes > show > show-hidden
+    View > Visibility > Methods and dunder attributes > hide > hide-hidden
+    Row > Dive into > open-row
+    System > Python > import library > import-python
+    System > Python > current sheet > pyobj-sheet
+    System > Python > current row > pyobj-row
+    System > Python > current cell > pyobj-cell
+    System > Python > expression > pyobj-expr
+    System > Python > exec() > exec-python
+''')
